@@ -1,9 +1,17 @@
 #!/usr/bin/python
+#
+# SPDX-License-Identifier: GPL-2.0-only
+# (c) 2023 Gerd Hoffmann
+#
 """ dump content of ovmf firmware volumes """
 import sys
+import json
 import lzma
 import struct
-import optparse
+import hashlib
+import argparse
+import tempfile
+import subprocess
 import collections
 
 from virt.firmware.efi import guids
@@ -269,15 +277,18 @@ class Edk2FfsFile(Edk2CommonBase):
         0x06 : 'peim',
         0x07 : 'driver',
         0x09 : 'application',
-        0x0a : 'smm',
+        0x0a : 'mm',
         0x0b : 'fw-volume',
-        0x0d : 'smm-core',
+        0x0d : 'mm-core',
+        0x0e : 'mm-standalone',
+        0x0f : 'mm-standalone-core',
         0xf0 : 'padding',
     }
     sectiontypes = (
         0x03, 0x04, 0x05,
         0x06, 0x07, 0x08, 0x09,
         0x0a, 0x0b, 0x0d,
+        0x0e, 0x0f,
     )
 
     def __init__(self, data = None):
@@ -323,6 +334,7 @@ class Edk2FfsFile(Edk2CommonBase):
         return f'ffsfile={guids.name(self.guid)} size=0x{self.tlen:x} type={self.fmt_type()}'
 
 
+# pylint: disable=too-many-instance-attributes
 class Edk2Variable(Edk2CommonBase):
     """ edk2 (non-volatile) variable """
 
@@ -333,19 +345,23 @@ class Edk2Variable(Edk2CommonBase):
         self.blob = b''
         self.state = 0
         self.attr = 0
+        self.nsize = 0
+        self.dsize = 0
         if data:
             self.parse(data)
 
     def parse(self, data):
-        (magic, self.state, self.attr, count, time, pk, nsize, dsize, guid) = \
+        (magic, self.state, self.attr, count, time, pk, self.nsize, self.dsize, guid) = \
             struct.unpack_from("<HBxLQ16sLLL16s", data)
         if magic != 0x55aa:
             raise ValueError('end of variable list')
         self.guid = guids.parse_bin(guid, 0)
-        self.name = ucs16.from_ucs16(data, 60)
-        self.blob = data [ 60 + nsize :
-                           60 + nsize + dsize ]
-        self.tlen = 60 + nsize + dsize
+        self.name = ucs16.from_ucs16(data [ 60 : 60 + self.nsize ], 0)
+        self.blob = data [ 60 + self.nsize :
+                           60 + self.nsize + self.dsize ]
+        self.tlen = 60 + self.nsize + self.dsize
+        if len(data) < self.tlen:
+            raise ValueError('invalid variable entry')
 
         if self.state == 0x3f:
             var = efivar.EfiVar(self.name,
@@ -359,14 +375,20 @@ class Edk2Variable(Edk2CommonBase):
                 self.append(f'blob: {len(self.blob)} bytes')
 
     def fmt_state(self):
+        if self.state == 0x7f:
+            return '(header-ok)'
         if self.state == 0x3f:
-            return None
+            return '(ok)'
+        if self.state == 0x3e:
+            return '(del-in-progress)'
         if self.state == 0x3c:
-            return '(deleted)'
+            return '(deleted)' # normal delete process
+        if self.state == 0x3d:
+            return '(deleted)' # without 'del-in-progress' step
         return f'state=0x{self.state:x}'
 
     def __str__(self):
-        ret = f'variable={guids.name(self.guid)} size=0x{self.tlen:x}'
+        ret = f'variable={guids.name(self.guid)} nsize=0x{self.nsize:x} dsize=0x{self.dsize:x}'
         ret += f' attr=0x{self.attr:x} name={self.name}'
         state = self.fmt_state()
         if state:
@@ -396,7 +418,7 @@ class Edk2NvData(Edk2CommonBase):
             try:
                 var = Edk2Variable(data [ self.used : ])
             except (ValueError, struct.error) as err:
-                self.append(f'{err}')
+                self.append(f'{err} at offset 0x{self.used:x}')
                 return
             self.used += var.size()
             self.align(4)
@@ -414,12 +436,14 @@ class Edk2Volume(Edk2CommonBase):
     def __init__(self, data = None, offset = 0):
         super().__init__()
         self.guid   = None
+        self.name   = None
         self.attr   = 0
         self.xoff   = 0
         self.rev    = 0
         self.blocks = 0
         self.bsize  = 0
         self.offset = offset
+        self.blob   = None
         if data:
             self.parse(data)
 
@@ -428,6 +452,9 @@ class Edk2Volume(Edk2CommonBase):
          csum, self.xoff, self.rev, self.blocks, self.bsize) = \
             struct.unpack_from('<16x16sQLLHHHxBLL', data)
         self.guid = guids.parse_bin(guid, 0)
+        self.blob = data [ 0 : self.tlen ]
+        if self.xoff:
+            self.name = guids.parse_bin(data, self.xoff)
 
         if self.is_ffs():
             self.parse_ffs(data [ self.hlen : self.tlen ])
@@ -455,11 +482,14 @@ class Edk2Volume(Edk2CommonBase):
         return str(self.guid) == guids.NvData
 
     def __str__(self):
-        return (f'volume={guids.name(self.guid)} '
-                f'offset=0x{self.offset:x} size=0x{self.tlen:x} '
-                f'hlen=0x{self.hlen:x} xoff=0x{self.xoff:x} '
-                f'rev={self.rev} blocks={self.blocks}*{self.bsize} '
-                f'used={self.used * 100 / self.tlen:.1f}%')
+        ret = (f'volume={guids.name(self.guid)} '
+               f'offset=0x{self.offset:x} size=0x{self.tlen:x} '
+               f'hlen=0x{self.hlen:x} xoff=0x{self.xoff:x} '
+               f'rev={self.rev} blocks={self.blocks}*{self.bsize} '
+               f'used={self.used * 100 / self.tlen:.1f}%')
+        if self.name:
+            ret += f' name={guids.name(self.name)}'
+        return ret
 
 
 class Edk2Capsule(Edk2CommonBase):
@@ -506,9 +536,12 @@ class Edk2Image(collections.UserList):
     def parse(self, data):
         pos = 0
         step = 1024
+        skips = 0
+        maxskips = 256
         while pos + 32 < len(data):
             (tlen, sig) = struct.unpack_from('<QL', data, pos + 32)
             if sig == 0x4856465f:
+                skips = 0
                 vol = Edk2Volume(data = data [ pos : ],
                                  offset = pos)
                 pos += vol.size()
@@ -517,11 +550,16 @@ class Edk2Image(collections.UserList):
 
             guid = guids.parse_bin(data, pos)
             if str(guid) == guids.SignedCapsule:
+                skips = 0
                 upd = Edk2Capsule(data [ pos : ] )
                 pos += upd.size()
                 self.append(upd)
                 continue
 
+            skips += 1
+            if skips > maxskips:
+                return
+
             pos = (pos + step) & ~(step-1)
             continue
 
@@ -559,50 +597,110 @@ def print_ovmf_meta(item, indent):
         return print_all(item, indent)
     return 0
 
-def print_tree(item, pfunc, indent = 0):
-    inc = pfunc(item, indent)
-    if isinstance(item, collections.UserList):
-        for i in list(item):
-            print_tree(i, pfunc, indent + inc)
+
+########################################################################
+# get hashes (for pcr precalculation)
+
+jsonimgs = []
+jsonvols = []
+
+# pylint: disable=unused-argument
+def get_volume_hashes(item, indent):
+    if isinstance(item, (Edk2Volume,)) and item.blob:
+        rec = {
+            'guid-type' : str(item.guid),
+            'sha256'    : hashlib.sha256(item.blob).hexdigest(),
+        }
+        if item.name:
+            rec['guid-name'] = str(item.name)
+            if str(item.name) == guids.OvmfPeiFv:
+                rec['name'] = "PEIFV"
+            if str(item.name) == guids.OvmfDxeFv:
+                rec['name'] = "DXEFV"
+        jsonvols.append(rec)
+    return 0
+
+def get_image_data(image):
+    jsonvols.clear()
+    walk_tree(image, get_volume_hashes)
+    rec = {
+        'image'   : image.name,
+        'volumes' : jsonvols.copy(),
+    }
+    jsonimgs.append(rec)
+
+def print_image_data():
+    print(json.dumps(jsonimgs, indent = 4, sort_keys = True))
 
 
 ########################################################################
 # main
 
+def unqcow2(filename, data):
+    (magic, version) = struct.unpack_from('>LL', data)
+    if magic != 0x514649fb:
+        return data
+    with tempfile.NamedTemporaryFile() as rawfile:
+        cmdline = [ 'qemu-img', 'convert',
+                    '-f', 'qcow2', '-O', 'raw',
+                    filename, rawfile.name ]
+        subprocess.run(cmdline, check = True)
+        filedata = rawfile.read()
+    return filedata
+
+def walk_tree(item, pfunc, indent = 0):
+    inc = pfunc(item, indent)
+    if isinstance(item, collections.UserList):
+        for i in list(item):
+            walk_tree(i, pfunc, indent + inc)
+
 def main():
-    parser = optparse.OptionParser()
-    parser.add_option('-i', '--input', dest = 'input', type = 'string',
-                      help = 'dump firmware volume FILE', metavar = 'FILE')
-    parser.add_option('--all', dest = 'fmt',
-                      action = 'store_const', const = 'all',
-                      help = 'print everything (default)')
-    parser.add_option('--volumes', dest = 'fmt',
-                      action = 'store_const', const = 'volumes',
-                      help = 'print firmware volumes')
-    parser.add_option('--modules', dest = 'fmt',
-                      action = 'store_const', const = 'modules',
-                      help = 'print included modules')
-    parser.add_option('--ovmf-meta', dest = 'fmt',
-                      action = 'store_const', const = 'ovmf-meta',
-                      help = 'decode ovmf metadata (in reset vector)')
-    (options, args) = parser.parse_args()
+    parser = argparse.ArgumentParser(
+        description = "Dump EFI firmware volumes and EFI variable stores.")
+
+    parser.add_argument('-i', '--input', dest = 'input',
+                        action = 'append', type = str,
+                        help = 'dump firmware volume FILE', metavar = 'FILE')
+    parser.add_argument('--all', dest = 'fmt',
+                        action = 'store_const', const = 'all',
+                        help = 'print everything (default)')
+    parser.add_argument('--volumes', dest = 'fmt',
+                        action = 'store_const', const = 'volumes',
+                        help = 'print firmware volumes')
+    parser.add_argument('--modules', dest = 'fmt',
+                        action = 'store_const', const = 'modules',
+                        help = 'print included modules')
+    parser.add_argument('--ovmf-meta', dest = 'fmt',
+                        action = 'store_const', const = 'ovmf-meta',
+                        help = 'decode ovmf metadata (in reset vector)')
+    parser.add_argument('--volume-hashes', dest = 'fmt',
+                        action = 'store_const', const = 'volume-hashes',
+                        help = 'print volume hashes')
+    options = parser.parse_args()
 
     if not options.input:
         print('ERROR: no input file specified (try -h for help)')
         sys.exit(1)
 
-    with open(options.input, 'rb') as f:
-        data = f.read()
-    image = Edk2Image(options.input, data)
+    for filename in options.input:
+        with open(filename, 'rb') as f:
+            data = f.read()
+        data = unqcow2(filename, data)
+        image = Edk2Image(filename, data)
 
-    if options.fmt == 'all' or options.fmt is None:
-        print_tree(image, print_all)
-    elif options.fmt == 'volumes':
-        print_tree(image, print_volumes)
-    elif options.fmt == 'modules':
-        print_tree(image, print_modules)
-    elif options.fmt == 'ovmf-meta':
-        print_tree(image, print_ovmf_meta)
+        if options.fmt == 'all' or options.fmt is None:
+            walk_tree(image, print_all)
+        elif options.fmt == 'volumes':
+            walk_tree(image, print_volumes)
+        elif options.fmt == 'modules':
+            walk_tree(image, print_modules)
+        elif options.fmt == 'ovmf-meta':
+            walk_tree(image, print_ovmf_meta)
+        elif options.fmt == 'volume-hashes':
+            get_image_data(image)
+
+    if options.fmt == 'volume-hashes':
+        print_image_data()
 
 if __name__ == '__main__':
     sys.exit(main())
